;----------------------------------------------------------------------------
;
; File:        _tms9900.asm
; Date:        23-Feb-1998
; Programmer:  Marc Rousseau
;
; Description:
;
; Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
;
; Revision History:
;
;----------------------------------------------------------------------------

	%INCLUDE	"tms9900.inc"

	DATASEG

STRUC	sOpCode
  .mnemonic	resb	8
  .opCode  	resw	1
  .mask		resw	1
  .format	resw	1
  .unused	resw	1
  .function	resd	1
  .clocks	resd	1
  .count	resd	1
ENDSTRUC

STRUC	sLookUp
  .opCode	resd	1
  .size		resd	1
ENDSTRUC

	CEXTERN	TimerHook

	EXTERN	intLevel
	CEXTERN	LookUp

	CGLOBAL	CpuMemory, MemFlags
	CGLOBAL InterruptFlag, WorkspacePtr, ProgramCounter, Status
	CGLOBAL InstructionCounter
	CGLOBAL ClockCycleCounter
	GLOBAL	CpuMemoryW, CpuMemoryB

InterruptFlag		db	-1

WorkspacePtr		dw	0
ProgramCounter		dw	0
Status			dw	0

InstructionCounter	dd	0
ClockCycleCounter	dd	0
	
stopFlag		dd	0
runFlag			dd	0

	UDATASEG

CpuMemoryW
CpuMemoryB
CpuMemory		resb	00010000h
MemFlags		resb	00010000h

	CODESEG

	CEXTERN	CallTrap
	CEXTERN	ExecuteGPL
	CEXTERN	InvalidOpcode
	EXTERN  ContextSwitch
	EXTERN  SetStatus
	EXTERN  GetStatus

	CGLOBAL	Run, Stop, Step, IsRunning

	GLOBAL	ExecuteInstruction
	GLOBAL	GetArgumentB, GetArgumentW
	GLOBAL	ReadMemoryB,   ReadMemoryW
	GLOBAL 	_ReadMemoryB,  _ReadMemoryW
	GLOBAL	WriteMemoryB,  WriteMemoryW
	GLOBAL	_WriteMemoryB, _WriteMemoryW

GetArgumentB:

	mov	ecx, eax
	and	ecx, 0000000Fh
	lea	ecx, [ _WP_ + ecx * 2 ]
	test	eax, 00000030h
	je	.done
	mov     edx, ecx
	movzx	ecx, word [ CpuMemoryW + ecx ]
	test	eax, 00000020h
	je	.done
	test	eax, 00000010h
	je	.indexed
	inc	word [ CpuMemoryW + edx ]
	jmp	.done

.indexed:
	cmp	_WP_, edx
	sbb	edx, edx
	and	ecx, edx
	add	cx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
.done:
	xor	ecx, 1
	ret

GetArgumentW:

	mov	ecx, eax
	and	ecx, 0000000Fh
	lea	ecx, [ _WP_ + ecx * 2 ]
	test	eax, 00000030h
	je	.done
	mov     edx, ecx
	movzx	ecx, word [ CpuMemoryW + ecx ]
	test	eax, 00000020h
	je	.done
	test	eax, 00000010h
	je	.indexed
 	add	word [ CpuMemoryW + edx ], 2
	jmp	.done

.indexed:
	cmp	_WP_, edx
	sbb	edx, edx
	and	ecx, edx
	add	cx, [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
.done:
	and	ecx, 0000FFFEh
	ret


ReadMemoryB:

	movsx	ebx, byte [ CpuMemoryB + ecx ]
	ret

ReadMemoryW:

	movsx	ebx, word [ CpuMemoryW + ecx ]
	ret

_ReadMemoryB:

	movsx	ebx, byte [ CpuMemoryB + ecx ]
	test	byte [ MemFlags + ecx ], MEMFLG_READ
	je	.MemOK
	CallTrapByte	1
.MemOK:
	ret

_ReadMemoryW:

	movzx	ebx, word [ CpuMemoryW + ecx ]
	test	byte [ MemFlags + ecx ], MEMFLG_READ
	je	.MemOK
	CallTrapWord	1
	movzx	ebx, bx
.MemOK:
	ret

WriteMemoryB:

	mov	[ CpuMemoryB + ecx ], bl
	ret

WriteMemoryW:

	mov	[ CpuMemoryW + ecx ], bx
	ret

_WriteMemoryB:

	test	byte [ MemFlags + ecx ], MEMFLG_WRITE | MEMFLG_ROM
	je	.Done
	test	byte [ MemFlags + ecx ], MEMFLG_WRITE
	je	.MemOK
	CallTrapByte	0
.MemOK:
	test	byte [ MemFlags + ecx ], MEMFLG_ROM
	jne	.ROM
.Done:
	mov	[ CpuMemoryB + ecx ], bl
.ROM:
	ret

_WriteMemoryW:

	test	byte [ MemFlags + ecx ], MEMFLG_WRITE | MEMFLG_ROM
	je	.Done
	test	byte [ MemFlags + ecx ], MEMFLG_WRITE
	je	.MemOK
	CallTrapWord	0
.MemOK:
	test	byte [ MemFlags + ecx ], MEMFLG_ROM
	jne	.ROM
.Done:
	mov	[ CpuMemoryW + ecx ], bx
.ROM:
	ret

ExecuteInstruction:

	mov	ebx, eax
	shr	ebx, 12
	lea	ebx, [ LookUp + ebx * 8 ]
	mov	edi, [ ebx + sLookUp.opCode ]
	mov	ecx, [ ebx + sLookUp.size ]
	jecxz	.Match
	inc     ecx
.Retry:
	movzx	ebx, word [ edi + sOpCode.mask ]
	and	ebx, eax
	cmp	bx, [ edi + sOpCode.opCode ]
	je	.Match
	add	edi, sOpCode_size
	loop	.Retry
	call	InvalidOpcode
	ret
.Match:
	mov	ebx, [ edi + sOpCode.clocks ]
	add	[ ClockCycleCounter ], ebx
	inc	dword [ edi + sOpCode.count ]
	call	[ edi + sOpCode.function ]
	ret

Step:

	inc	dword [ runFlag ]
	pushad
	movzx	ebp, word [ WorkspacePtr ]
	movzx	esi, word [ ProgramCounter ]
	movzx	edx, word [ Status ]

	call	SetStatus

	call	CheckInterrupt
	movzx	eax, word [ CpuMemoryW + _PC_ ]
	add	_PCw_, 2
	call	ExecuteInstruction
	inc	dword [ InstructionCounter ]

        cmp	dword [ TimerHook ], 0
        je	.Continue
        mov	eax, [ InstructionCounter ]
        and	eax, 0FFh
        jne	.Continue
        pushad
        call	[ TimerHook ]
        popad

.Continue:

	call	GetStatus

	mov	[ WorkspacePtr ], bp
	mov	[ ProgramCounter ], si
	mov	[ Status ], dx
	popad
	dec	dword [ runFlag ]
	xor	eax,eax
	cmp	dword [ stopFlag ],0
        je	.Done
	dec	dword [ stopFlag ]
        inc	eax
.Done:
	ret

Run:

	inc	dword [ runFlag ]
	pushad
	movzx	ebp, word [ WorkspacePtr ]
	movzx	esi, word [ ProgramCounter ]
	movzx	edx, word [ Status ]

	call	SetStatus

.Forever:

	call	CheckInterrupt
;	cmp	_PCw_, 0070h
;	jne	.NormalCode
;	pushad
;	call	ExecuteGPL
;	cmp	eax, 0
;	je	.NoGPL
;	popad
;	jmp	.CheckTimer

.NoGPL:

;	popad

.NormalCode:

	movzx	eax, word [ CpuMemoryW + _PC_ ]
	mov	[ ProgramCounter ], _PCw_
	add	_PCw_, 2
	call	ExecuteInstruction
	inc	dword [ InstructionCounter ]

.CheckTimer:

        cmp	dword [ TimerHook ], 0
        je	.Continue
        mov	eax, [ InstructionCounter ]
        and	eax, 0FFh
        jne	.Continue
        pushad
        call	[ TimerHook ]
        popad

.Continue:

	cmp	dword [ stopFlag ], 0
	je	.Forever

	call	GetStatus

	mov	[ WorkspacePtr ], bp
	mov	[ ProgramCounter ], si
	mov	[ Status ], dx
	dec	dword [ stopFlag ]
	popad
	dec	dword [ runFlag ]
	ret

Stop:

	inc	dword [ stopFlag ]
	ret

IsRunning:

	mov	eax, [ runFlag ]
	ret

CheckInterrupt:

	movzx   ecx, byte [ InterruptFlag ]
	cmp     cl, [ intLevel ]
	ja      .Done

	mov     byte [ InterruptFlag ], -1
	movzx   eax, word [ CpuMemoryW + ecx * 4 ]
	movzx   ebx, word [ CpuMemoryW + ecx * 4 + 2 ]
	call    ContextSwitch

	cmp     cl, 0
	je      .Done
	dec     cl
	mov     [ intLevel ], cl

.Done:
	ret

	END
